@@ -222,15 +222,18 @@ static char *ask_new_field(struct chfn_control *ctl, const char *question,
 			   char *def_val)
 {
 	int len;
-	char *buf;
+	char *buf = NULL; /* leave initialized to NULL or getline segfaults */
 	size_t dummy = 0;
 
 	if (!def_val)
 		def_val = "";
+
 	while (true) {
 		printf("%s [%s]:", question, def_val);
 		__fpurge(stdin);
+
 		putchar(' ');
+		fflush(stdout);
 
 		if (getline(&buf, &dummy, stdin) < 0)
 			errx(EXIT_FAILURE, _("Aborted."));
