@@ -37,7 +37,7 @@
 #include "strutils.h"
 #include "optutils.h"
 #include "xalloc.h"
-#include "procutils.h"
+#include "procfs.h"
 #include "ipcutils.h"
 #include "timeutils.h"
 
@@ -129,6 +129,7 @@ struct lsipc_control {
 	int outmode;
 	unsigned int noheadings : 1,		/* don't print header line */
 		     notrunc : 1,		/* don't truncate columns */
+		     shellvar : 1,              /* use shell compatible colum names */
 		     bytes : 1,			/* SIZE in bytes */
 		     numperms : 1,		/* numeric permissions */
 		     time_mode : 2;
@@ -315,6 +316,8 @@ static void __attribute__((__noreturn__)) usage(void)
 	fputs(_(" -P, --numeric-perms      print numeric permissions (PERMS column)\n"), out);
 	fputs(_(" -r, --raw                display in raw mode\n"), out);
 	fputs(_(" -t, --time               show attach, detach and change times\n"), out);
+	fputs(_(" -y, --shell              use column names to be usable as shell variable identifiers\n"), out);
+
 
 	fputs(USAGE_SEPARATOR, out);
 	printf(USAGE_HELP_OPTIONS(26));
@@ -352,6 +355,8 @@ static struct libscols_table *new_table(struct lsipc_control *ctl)
 
 	if (ctl->noheadings)
 		scols_table_enable_noheadings(table, 1);
+	if (ctl->shellvar)
+		scols_table_enable_shellvar(table, 1);
 
 	switch(ctl->outmode) {
 	case OUT_NEWLINE:
@@ -469,8 +474,9 @@ static char *make_time(int mode, time_t time)
 	return xstrdup(buf);
 }
 
-static void global_set_data(struct libscols_table *tb, const char *resource,
-			    const char *desc, uintmax_t used, uintmax_t limit, int usage)
+static void global_set_data(struct lsipc_control *ctl, struct libscols_table *tb,
+			    const char *resource, const char *desc, uintmax_t used,
+			    uintmax_t limit, int usage, int byte_unit)
 {
 	struct libscols_line *ln;
 	size_t n;
@@ -492,7 +498,10 @@ static void global_set_data(struct libscols_table *tb, const char *resource,
 			break;
 		case COL_USED:
 			if (usage) {
-				xasprintf(&arg, "%ju", used);
+				if (!byte_unit || ctl->bytes)
+					xasprintf(&arg, "%ju", used);
+				else
+					arg = size_to_human_string(SIZE_SUFFIX_1LETTER, used);
 				rc = scols_line_refer_data(ln, n, arg);
 			} else
 				rc = scols_line_set_data(ln, n, "-");
@@ -505,7 +514,10 @@ static void global_set_data(struct libscols_table *tb, const char *resource,
 				rc = scols_line_set_data(ln, n, "-");
 			break;
 		case COL_LIMIT:
-			xasprintf(&arg, "%ju", limit);
+			if (!byte_unit || ctl->bytes)
+				xasprintf(&arg, "%ju", limit);
+			else
+				arg = size_to_human_string(SIZE_SUFFIX_1LETTER, limit);
 			rc = scols_line_refer_data(ln, n, arg);
 			break;
 		}
@@ -687,7 +699,7 @@ static void do_sem(int id, struct lsipc_control *ctl, struct libscols_table *tb)
 					break;
 
 				/* COMMAND */
-				arg = proc_get_command(e->pid);
+				arg = pid_get_cmdline(e->pid);
 				rc = scols_line_refer_data(sln, 5, arg);
 				if (rc)
 					break;
@@ -703,7 +715,7 @@ static void do_sem(int id, struct lsipc_control *ctl, struct libscols_table *tb)
 	ipc_sem_free_info(semds);
 }
 
-static void do_sem_global(struct libscols_table *tb)
+static void do_sem_global(struct lsipc_control *ctl, struct libscols_table *tb)
 {
 	struct sem_data *semds, *semdsp;
 	struct ipc_limits lim;
@@ -719,11 +731,11 @@ static void do_sem_global(struct libscols_table *tb)
 		ipc_sem_free_info(semds);
 	}
 
-	global_set_data(tb, "SEMMNI", _("Number of semaphore identifiers"), nsets, lim.semmni, 1);
-	global_set_data(tb, "SEMMNS", _("Total number of semaphores"), nsems, lim.semmns, 1);
-	global_set_data(tb, "SEMMSL", _("Max semaphores per semaphore set."), 0, lim.semmsl, 0);
-	global_set_data(tb, "SEMOPM", _("Max number of operations per semop(2)"), 0, lim.semopm, 0);
-	global_set_data(tb, "SEMVMX", _("Semaphore max value"), 0, lim.semvmx, 0);
+	global_set_data(ctl, tb, "SEMMNI", _("Number of semaphore identifiers"), nsets, lim.semmni, 1, 0);
+	global_set_data(ctl, tb, "SEMMNS", _("Total number of semaphores"), nsems, lim.semmns, 1, 0);
+	global_set_data(ctl, tb, "SEMMSL", _("Max semaphores per semaphore set."), 0, lim.semmsl, 0, 0);
+	global_set_data(ctl, tb, "SEMOPM", _("Max number of operations per semop(2)"), 0, lim.semopm, 0, 0);
+	global_set_data(ctl, tb, "SEMVMX", _("Semaphore max value"), 0, lim.semvmx, 0, 0);
 }
 
 static void do_msg(int id, struct lsipc_control *ctl, struct libscols_table *tb)
@@ -826,7 +838,10 @@ static void do_msg(int id, struct lsipc_control *ctl, struct libscols_table *tb)
 							  (time_t)msgdsp->q_ctime));
 				break;
 			case COL_USEDBYTES:
-				xasprintf(&arg, "%ju", msgdsp->q_cbytes);
+				if (ctl->bytes)
+					xasprintf(&arg, "%ju", msgdsp->q_cbytes);
+				else
+					arg = size_to_human_string(SIZE_SUFFIX_1LETTER, msgdsp->q_cbytes);
 				rc = scols_line_refer_data(ln, n, arg);
 				break;
 			case COL_MSGS:
@@ -865,7 +880,7 @@ static void do_msg(int id, struct lsipc_control *ctl, struct libscols_table *tb)
 }
 
 
-static void do_msg_global(struct libscols_table *tb)
+static void do_msg_global(struct lsipc_control *ctl, struct libscols_table *tb)
 {
 	struct msg_data *msgds, *msgdsp;
 	struct ipc_limits lim;
@@ -880,9 +895,9 @@ static void do_msg_global(struct libscols_table *tb)
 		ipc_msg_free_info(msgds);
 	}
 
-	global_set_data(tb, "MSGMNI", _("Number of message queues"), msgqs, lim.msgmni, 1);
-	global_set_data(tb, "MSGMAX", _("Max size of message (bytes)"),	0, lim.msgmax, 0);
-	global_set_data(tb, "MSGMNB", _("Default max size of queue (bytes)"), 0, lim.msgmnb, 0);
+	global_set_data(ctl, tb, "MSGMNI", _("Number of message queues"), msgqs, lim.msgmni, 1, 0);
+	global_set_data(ctl, tb, "MSGMAX", _("Max size of message (bytes)"),	0, lim.msgmax, 0, 1);
+	global_set_data(ctl, tb, "MSGMNB", _("Default max size of queue (bytes)"), 0, lim.msgmnb, 0, 1);
 }
 
 
@@ -1049,7 +1064,7 @@ static void do_shm(int id, struct lsipc_control *ctl, struct libscols_table *tb)
 				rc = scols_line_refer_data(ln, n, arg);
 				break;
 			case COL_COMMAND:
-				arg = proc_get_command(shmdsp->shm_cprid);
+				arg = pid_get_cmdline(shmdsp->shm_cprid);
 				rc = scols_line_refer_data(ln, n, arg);
 				break;
 			}
@@ -1063,7 +1078,7 @@ static void do_shm(int id, struct lsipc_control *ctl, struct libscols_table *tb)
 	ipc_shm_free_info(shmds);
 }
 
-static void do_shm_global(struct libscols_table *tb)
+static void do_shm_global(struct lsipc_control *ctl, struct libscols_table *tb)
 {
 	struct shm_data *shmds, *shmdsp;
 	uint64_t nsegs = 0, sum_segsz = 0;
@@ -1079,10 +1094,10 @@ static void do_shm_global(struct libscols_table *tb)
 		ipc_shm_free_info(shmds);
 	}
 
-	global_set_data(tb, "SHMMNI", _("Shared memory segments"), nsegs, lim.shmmni, 1);
-	global_set_data(tb, "SHMALL", _("Shared memory pages"), sum_segsz / getpagesize(), lim.shmall, 1);
-	global_set_data(tb, "SHMMAX", _("Max size of shared memory segment (bytes)"), 0, lim.shmmax, 0);
-	global_set_data(tb, "SHMMIN", _("Min size of shared memory segment (bytes)"), 0, lim.shmmin, 0);
+	global_set_data(ctl, tb, "SHMMNI", _("Shared memory segments"), nsegs, lim.shmmni, 1, 0);
+	global_set_data(ctl, tb, "SHMALL", _("Shared memory pages"), sum_segsz / getpagesize(), lim.shmall, 1, 0);
+	global_set_data(ctl, tb, "SHMMAX", _("Max size of shared memory segment (bytes)"), 0, lim.shmmax, 0, 1);
+	global_set_data(ctl, tb, "SHMMIN", _("Min size of shared memory segment (bytes)"), 0, lim.shmmin, 0, 1);
 }
 
 int main(int argc, char *argv[])
@@ -1122,6 +1137,7 @@ int main(int argc, char *argv[])
 		{ "time",           no_argument,	NULL, 't' },
 		{ "time-format",    required_argument,	NULL, OPT_TIME_FMT },
 		{ "version",        no_argument,	NULL, 'V' },
+		{ "shell",          no_argument,	NULL, 'y' },
 		{NULL, 0, NULL, 0}
 	};
 
@@ -1143,7 +1159,7 @@ int main(int argc, char *argv[])
 
 	scols_init_debug(0);
 
-	while ((opt = getopt_long(argc, argv, "bceghi:Jlmno:PqrstV", longopts, NULL)) != -1) {
+	while ((opt = getopt_long(argc, argv, "bceghi:Jlmno:PqrstVy", longopts, NULL)) != -1) {
 
 		err_exclusive_options(opt, longopts, excl, excl_st);
 
@@ -1232,6 +1248,9 @@ int main(int argc, char *argv[])
 		case 'c':
 			show_creat = 1;
 			break;
+		case 'y':
+			ctl->shellvar = 1;
+			break;
 
 		case 'h':
 			usage();
@@ -1311,19 +1330,19 @@ int main(int argc, char *argv[])
 
 	if (msg) {
 		if (global)
-			do_msg_global(tb);
+			do_msg_global(ctl, tb);
 		else
 			do_msg(id, ctl, tb);
 	}
 	if (shm) {
 		if (global)
-			do_shm_global(tb);
+			do_shm_global(ctl ,tb);
 		else
 			do_shm(id, ctl, tb);
 	}
 	if (sem) {
 		if (global)
-			do_sem_global(tb);
+			do_sem_global(ctl, tb);
 		else
 			do_sem(id, ctl, tb);
 	}
