@@ -321,9 +321,9 @@ main(int argc, char *argv[])
 				execl(shell, shname, "-i", (char *)NULL);
 		} else {
 			if (command)
-				execlp(shname, "-c", command, (char *)NULL);
+				execlp(shname, shname, "-c", command, (char *)NULL);
 			else
-				execlp(shname, "-i", (char *)NULL);
+				execlp(shname, shname, "-i", (char *)NULL);
 		}
 		err(EXIT_FAILURE, "failed to execute %s", shell);
 		break;
